# coding: utf-8
from collections import Iterable
from collections import defaultdict

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db.transaction import commit_on_success
from django.utils.importlib import import_module
from m3.actions import ControllerCache
from redlock import RedLock

from educommon import ioc

from .models import Permission
from .models import Role
from .models import UserRole


def _get_lock():
    u"""Возвращает блокировку Redis в виде менеджера контекста.

    :rtype: redlock.RedLock
    """
    redis_connection_details = [
        dict(
            host=settings.REDIS_HOST,
            port=settings.REDIS_PORT,
            db=settings.REDIS_DB
        ),
    ]

    result = RedLock('rbac_lock', redis_connection_details)

    return result


def _get_handler(handler):
    u"""Возвращает функцию-обработчик правила."""
    module, handler = handler.rsplit('.', 1)
    module = __import__(module, fromlist=[handler])
    handler = getattr(module, handler)
    return handler


def _get_app_permissions_modules():
    u"""Возвращает модули permissions из приложений системы.

    :rtype: generator
    """
    for app_name in settings.INSTALLED_APPS:
        try:
            yield import_module('.permissions', app_name)
        except ImportError as error:
            if 'No module named' not in error.args[0]:
                raise
            continue


def _get_actions():
    u"""Возвращает экшены системы, доступ к которым нужно проверять.

    :rtype: generator
    """
    for controller in ControllerCache.get_controllers():
        for pack in controller.get_packs():
            for action in pack.actions:
                if pack.need_check_permission or action.need_check_permission:
                    yield action


def _set_permission(name, title, description):
    u"""Сохраняет разрешение в БД.

    Если разрешение отсутствует в БД, то создает его. Иначе обновляет его
    обработчик, название и дескриптор.
    """
    try:
        permission = Permission.objects.get(name=name)
    except Permission.DoesNotExist:
        permission = Permission(
            name=name,
            title=title,
            description=description,
        )
        permission.full_clean()
        permission.save()
    else:
        if (
            permission.title != title
            or permission.description != description
        ):
            permission.title = title
            permission.description = description
            permission.full_clean()
            permission.save()

    return permission


class RBACManager(object):

    u"""Менеджер системы авторизации RBAC."""

    def __init__(self):
        self._initialized = False
        self.permissions_by_name = {}
        self.permission_rules = defaultdict(set)

    def _collect_permissions(self, permissions_modules):
        u"""Сбор разрешений системы."""
        self.permissions_by_name.clear()

        # Сбор названий разрешений всей системы
        for action in _get_actions():
            if action.sub_permissions:
                for sub_perm, title in action.sub_permissions.iteritems():
                    assert title, (action.parent, action, sub_perm)
                    perm_name = action.get_perm_code(sub_perm)
                    self.permissions_by_name[perm_name] = [title, None]
            else:
                title = action.parent.sub_permissions.get(action.perm_code)
                perm_name = action.get_perm_code()
                self.permissions_by_name[perm_name] = [title, None]

        # Заполнение параметров разрешений
        for module in permissions_modules:
            for name, title, description in getattr(module, 'permissions', []):
                assert name in self.permissions_by_name, (
                    'Permission {} not found'.format(name)
                )

                _title, _description = self.permissions_by_name[name]
                self.permissions_by_name[name][:] = [
                    title or _title,
                    description or _description,
                ]

    def _collect_rules(self, permissions_modules):
        u"""Сбор обработчиков правил для разрешений системы."""
        self.permission_rules.clear()

        for module in permissions_modules:
            for perm, handlers in getattr(module, 'rules', {}).iteritems():
                if not isinstance(handlers, Iterable):
                    handlers = [handlers]

                for handler in handlers:
                    if isinstance(handler, basestring):
                        handler = _get_handler(handler)
                    assert callable(handler), handler
                    self.permission_rules[perm].add(handler)

    @commit_on_success
    def init(self, update_db=True):
        u"""Инициализация системы авторизации.

        1. Загружает из приложений системы списки правил и разрешений. Их поиск
           осуществляется в модуле ``permissions``.
        2. Для каждого правила и разрешения создает/обновляет запись в БД.
        3. Объекты модели ``Permission`` сохраняет в словаре
           ``self.permissions_by_name``.

        :param bool update_db: Определяет необходимость синхронизации прав
            доступа системы с БД. Если синхронизация с БД не выполняется, то в
            словаре ``permissions_by_name`` будут параметры разрешений, а не
            сами разрешения!
        """
        ControllerCache.populate()

        modules = list(_get_app_permissions_modules())

        # Сбор разрешений системы
        self._collect_permissions(modules)

        # Сбор обработчиков правил для разрешений системы
        self._collect_rules(modules)

        if update_db:
            # Обновление прав доступа в БД
            with _get_lock():
                for params in self.permissions_by_name.iteritems():
                    name, (title, description) = params
                    permission = _set_permission(name, title, description)
                    self.permissions_by_name[name] = permission

    def has_perm(self, action, request):
        u"""Проверяет наличие у текущего пользователя наличие разрешения.

        Если у пака need_check_permission равен True, то проверка прав доступа
        будет выполняться для всех экшенов этого пака вне зависимости от
        значения параметра need_check_permission у экшенов. Также проверка
        наличия разрешений будет выполняться, если у пака значение параметра
        need_check_permission равно False, но у экшена оно равно True. В
        остальных случаях метод сразу возвращает True.

        :param request: HTTP-запрос.
        :type request: django.http.HttpRequest

        :param str permission_name: Название разрешения (Permission.name).

        :rtype: bool
        """
        if (not action.parent.need_check_permission and
                not action.need_check_permission):
            return True

        user = ioc.get('get_current_user')(request)
        if user is None:
            return False

        # Роли, назначенные непосредственно пользователю
        user_direct_roles = Role.objects.filter(
            pk__in=UserRole.objects.filter(
                content_type=ContentType.objects.get_for_model(user),
                object_id=user.id,
            ).values('role')
        )

        # Все роли пользователя с учетом подчиненных
        user_roles = set()
        for role in user_direct_roles:
            user_roles.add(role)
            user_roles.update(role.subroles)

        # Имя разрешения
        if action.sub_permissions:
            permission_names = [
                action.get_perm_code(sub_perm)
                for sub_perm in action.sub_permissions
            ]
        else:
            permission_names = [action.get_perm_code()]

        # M2M-модель, связывающая Role и Permission
        m2m_model = getattr(Role, '_meta').get_field('permissions').rel.through

        # Все разрешения пользователя
        user_permissions = Permission.objects.filter(
            pk__in=m2m_model.objects.filter(
                role__in=user_roles,
                permission__name__in=permission_names,
            ).distinct('permission').values('permission'),
        ).values_list('name', flat=True)

        for permission_name in user_permissions:
            if permission_name not in self.permission_rules:
                # Для разрешения не определено правил, значит достаточно
                # только наличия у пользователя разрешения как такового.
                return True

            for handler in self.permission_rules[permission_name]:
                if handler(action=action, request=request, user=user):
                    return True

        return False


rbac = RBACManager()
