# coding: utf-8
import json

from django import http
from django.contrib import auth
from django.http import HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from m3.actions.results import PreJsonResult
from m3_ext.ui.shortcuts import MessageBox
from objectpack.actions import BaseAction, BasePack

from . import checkers
from . import const


class AuthPack(BasePack):

    u"""Пак аутентификации."""

    url = const.AUTH_PACK_URL

    login_checker = checkers.DefaultLoginChecker()

    def __init__(self):
        super(AuthPack, self).__init__()
        self.login_page_action = LoginPageAction()
        self.login_action = LoginAction()

        self.logout_confirm_action = LogoutConfirmAction()
        self.logout_action = LogoutAction()

        self.actions.extend((
            self.login_page_action,
            self.logout_confirm_action,
            self.login_action,
            self.logout_action
        ))

    def declare_context(self, action):
        ctx = super(AuthPack, self).declare_context(action)
        if action is self.login_action:
            ctx['login'] = dict(type=unicode)
            ctx['password'] = dict(type=unicode)
        return ctx

    def get_login_url(self):
        return self.login_action.get_absolute_url()

    def get_logout_url(self):
        return self.logout_action.get_absolute_url()


class LoginPageAction(BaseAction):

    u"""Экшн отображает страницу входа."""

    url = const.LOGIN_PAGE_URL
    template_file_name = 'simple_auth/login_page.html'

    def get_template_context(self, request, context):
        return dict(
            login_url=self.parent.get_login_url()
        )

    def run(self, request, context):
        return render(
            request, self.template_file_name,
            self.get_template_context(request, context)
        )


class LoginAction(BaseAction):

    u"""Проверяет данные и выполняет вход."""

    url = '/login'

    def run(self, request, context):

        username = context.login
        password = context.password

        err_msg = self.parent.login_checker.check(request, username, password)

        # По-умолчанию создаем неудачный ответ с сообщением об ошибке:
        result = HttpResponse(
            json.dumps({
                'success': False,
                'redirect': '',
                'message': err_msg or ''
            })
        )

        if not err_msg:
            # Выполняем действия по аутентификации и логину пользователя:
            result = HttpResponse('{"success": true, "redirect": "/"}')
            # Просто аутентифицируем штатным образом:
            user = auth.authenticate(username=username, password=password)
            # Залогиним пользователя:
            auth.login(request, user)

        if request.is_ajax():
            return result
        else:
            return HttpResponseRedirect('/')


class LogoutConfirmAction(BaseAction):

    u"Экшн для отображения подтверждения выхода."

    url = '/logout-confirm'

    def run(self, request, context):
        msg_box = MessageBox(
            u'', u'Вы действительно хотите выйти из системы?',
            MessageBox.ICON_QUESTION, MessageBox.BTN_YESNO)

        msg_box.handler_yes = u"""
        Ext.Ajax.request({
            url: '%(url)s',
            params: {'confirm': true},
            success: function(response){
                var json = Ext.util.JSON.decode(response.responseText);
                window.location = json.redirect ? json.redirect : '/';
            }
        });
        """ % {'url': self.parent.logout_action.get_absolute_url()}
        return http.HttpResponse(msg_box.get_script())


class LogoutAction(BaseAction):

    u"Экшн выполняет выход."

    url = '/logout'

    def run(self, request, context):

        auth.logout(request)

        if request.is_ajax():
            return PreJsonResult({})
        else:
            return http.HttpResponseRedirect('/')
